# Copyright 2013-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import asyncio
import os
import pathlib
import time
import unittest
from test import IntegrationTest
from test.unified_format import generate_test_classes, get_test_path
from test.utils_shared import CMAPListener
from typing import Any, Optional

import pytest

from pymongo import MongoClient
from pymongo.driver_info import DriverInfo
from pymongo.monitoring import ConnectionClosedEvent

try:
    from mockupdb import MockupDB, OpMsgReply

    _HAVE_MOCKUPDB = True
except ImportError:
    _HAVE_MOCKUPDB = False

pytestmark = pytest.mark.mockupdb

_IS_SYNC = True

# Generate unified tests.
globals().update(generate_test_classes(get_test_path("handshake", "unified"), module=__name__))


def _get_handshake_driver_info(request):
    assert "client" in request
    return request["client"]


class TestClientMetadataProse(IntegrationTest):
    def setUp(self):
        super().setUp()
        self.server = MockupDB()
        self.handshake_req = None

        def respond(r):
            if "ismaster" in r:
                # then this is a handshake request
                self.handshake_req = r
            return r.reply(OpMsgReply(maxWireVersion=13))

        self.server.autoresponds(respond)
        self.server.run()
        self.addCleanup(self.server.stop)

    def send_ping_and_get_metadata(
        self, client: MongoClient, is_handshake: bool
    ) -> tuple[str, Optional[str], Optional[str], dict[str, Any]]:
        # reset if handshake request
        if is_handshake:
            self.handshake_req: Optional[dict] = None

        client.admin.command("ping")
        metadata = _get_handshake_driver_info(self.handshake_req)
        driver_metadata = metadata["driver"]
        name, version, platform = (
            driver_metadata["name"],
            driver_metadata["version"],
            metadata["platform"],
        )
        return name, version, platform, metadata

    def check_metadata_added(
        self,
        client: MongoClient,
        add_name: str,
        add_version: Optional[str],
        add_platform: Optional[str],
    ) -> None:
        # send initial metadata
        name, version, platform, metadata = self.send_ping_and_get_metadata(client, True)
        # wait for connection to become idle
        time.sleep(0.005)

        # add new metadata
        client.append_metadata(DriverInfo(add_name, add_version, add_platform))
        new_name, new_version, new_platform, new_metadata = self.send_ping_and_get_metadata(
            client, True
        )
        if add_name is not None and add_name.lower() in name.lower().split("|"):
            self.assertEqual(name, new_name)
            self.assertEqual(version, new_version)
            self.assertEqual(platform, new_platform)
        else:
            self.assertEqual(new_name, f"{name}|{add_name}" if add_name is not None else name)
            self.assertEqual(
                new_version,
                f"{version}|{add_version}" if add_version is not None else version,
            )
            self.assertEqual(
                new_platform,
                f"{platform}|{add_platform}" if add_platform is not None else platform,
            )

        metadata.pop("driver")
        metadata.pop("platform")
        new_metadata.pop("driver")
        new_metadata.pop("platform")
        self.assertEqual(metadata, new_metadata)

    def test_append_metadata(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
            driver=DriverInfo("library", "1.2", "Library Platform"),
        )
        self.check_metadata_added(client, "framework", "2.0", "Framework Platform")

    def test_append_metadata_platform_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
            driver=DriverInfo("library", "1.2", "Library Platform"),
        )
        self.check_metadata_added(client, "framework", "2.0", None)

    def test_append_metadata_version_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
            driver=DriverInfo("library", "1.2", "Library Platform"),
        )
        self.check_metadata_added(client, "framework", None, "Framework Platform")

    def test_append_metadata_platform_version_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
            driver=DriverInfo("library", "1.2", "Library Platform"),
        )
        self.check_metadata_added(client, "framework", None, None)

    def test_multiple_successive_metadata_updates(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string, maxIdleTimeMS=1, connect=False
        )
        client.append_metadata(DriverInfo("library", "1.2", "Library Platform"))
        self.check_metadata_added(client, "framework", "2.0", "Framework Platform")

    def test_multiple_successive_metadata_updates_platform_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
        )
        client.append_metadata(DriverInfo("library", "1.2", "Library Platform"))
        self.check_metadata_added(client, "framework", "2.0", None)

    def test_multiple_successive_metadata_updates_version_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
        )
        client.append_metadata(DriverInfo("library", "1.2", "Library Platform"))
        self.check_metadata_added(client, "framework", None, "Framework Platform")

    def test_multiple_successive_metadata_updates_platform_version_none(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
        )
        client.append_metadata(DriverInfo("library", "1.2", "Library Platform"))
        self.check_metadata_added(client, "framework", None, None)

    def test_doesnt_update_established_connections(self):
        listener = CMAPListener()
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
            driver=DriverInfo("library", "1.2", "Library Platform"),
            event_listeners=[listener],
        )

        # send initial metadata
        name, version, platform, metadata = self.send_ping_and_get_metadata(client, True)
        self.assertIsNotNone(name)
        self.assertIsNotNone(version)
        self.assertIsNotNone(platform)

        # add data
        add_name, add_version, add_platform = "framework", "2.0", "Framework Platform"
        client.append_metadata(DriverInfo(add_name, add_version, add_platform))
        # check new data isn't sent
        self.handshake_req: Optional[dict] = None
        client.admin.command("ping")
        self.assertIsNone(self.handshake_req)
        self.assertEqual(listener.event_count(ConnectionClosedEvent), 0)

    def test_duplicate_driver_name_no_op(self):
        client = self.rs_or_single_client(
            "mongodb://" + self.server.address_string,
            maxIdleTimeMS=1,
        )
        client.append_metadata(DriverInfo("library", "1.2", "Library Platform"))
        self.check_metadata_added(client, "framework", None, None)
        # wait for connection to become idle
        time.sleep(0.005)
        # add same metadata again
        self.check_metadata_added(client, "Framework", None, None)

    def test_handshake_documents_include_backpressure(self):
        # Create a `MongoClient` that is configured to record all handshake documents sent to the server as a part of
        # connection establishment.
        client = self.rs_or_single_client("mongodb://" + self.server.address_string)

        # Send a `ping` command to the server and verify that the command succeeds. This ensure that a connection is
        # established on all topologies.  Note: MockupDB only supports standalone servers.
        client.admin.command("ping")

        # Assert that for every handshake document intercepted:
        # the document has a field `backpressure` whose value is `true`.
        self.assertEqual(self.handshake_req["backpressure"], True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2020-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for caching OCSP responses."""

from __future__ import annotations

from collections import namedtuple
from datetime import datetime as _datetime
from datetime import timezone
from typing import TYPE_CHECKING, Any, Optional

from pymongo.lock import _create_lock

if TYPE_CHECKING:
    from cryptography.x509.ocsp import OCSPRequest, OCSPResponse


def _next_update(value: OCSPResponse) -> Optional[_datetime]:
    """Compat helper to return the response's next_update_utc."""
    # Added in cryptography 43.0.0.
    if hasattr(value, "next_update_utc"):
        return value.next_update_utc
    return value.next_update


def _this_update(value: OCSPResponse) -> Optional[_datetime]:
    """Compat helper to return the response's this_update_utc."""
    # Added in cryptography 43.0.0.
    if hasattr(value, "this_update_utc"):
        return value.this_update_utc
    return value.this_update


class _OCSPCache:
    """A cache for OCSP responses."""

    CACHE_KEY_TYPE = namedtuple(  # type: ignore
        "OcspResponseCacheKey",
        ["hash_algorithm", "issuer_name_hash", "issuer_key_hash", "serial_number"],
    )

    def __init__(self) -> None:
        self._data: dict[Any, OCSPResponse] = {}
        # Hold this lock when accessing _data.
        self._lock = _create_lock()

    def _get_cache_key(self, ocsp_request: OCSPRequest) -> CACHE_KEY_TYPE:
        return self.CACHE_KEY_TYPE(
            hash_algorithm=ocsp_request.hash_algorithm.name.lower(),
            issuer_name_hash=ocsp_request.issuer_name_hash,
            issuer_key_hash=ocsp_request.issuer_key_hash,
            serial_number=ocsp_request.serial_number,
        )

    def __setitem__(self, key: OCSPRequest, value: OCSPResponse) -> None:
        """Add/update a cache entry.

        'key' is of type cryptography.x509.ocsp.OCSPRequest
        'value' is of type cryptography.x509.ocsp.OCSPResponse

        Validity of the OCSP response must be checked by caller.
        """
        with self._lock:
            cache_key = self._get_cache_key(key)

            # As per the OCSP protocol, if the response's nextUpdate field is
            # not set, the responder is indicating that newer revocation
            # information is available all the time.
            next_update = _next_update(value)
            if next_update is None:
                self._data.pop(cache_key, None)
                return

            this_update = _this_update(value)
            if this_update is None:
                return
            now = _datetime.now(tz=timezone.utc)
            if this_update.tzinfo is None:
                # Make naive to match cryptography.
                now = now.replace(tzinfo=None)
            # Do nothing if the response is invalid.
            if not (this_update <= now < next_update):
                return

            # Cache new response OR update cached response if new response
            # has longer validity.
            cached_value = self._data.get(cache_key, None)
            if cached_value is None:
                self._data[cache_key] = value
                return
            cached_next_update = _next_update(cached_value)
            if cached_next_update is not None and cached_next_update < next_update:
                self._data[cache_key] = value

    def __getitem__(self, item: OCSPRequest) -> OCSPResponse:
        """Get a cache entry if it exists.

        'item' is of type cryptography.x509.ocsp.OCSPRequest

        Raises KeyError if the item is not in the cache.
        """
        with self._lock:
            cache_key = self._get_cache_key(item)
            value = self._data[cache_key]

            # Return cached response if it is still valid.
            this_update = _this_update(value)
            next_update = _next_update(value)
            assert this_update is not None
            assert next_update is not None
            now = _datetime.now(tz=timezone.utc)
            if this_update.tzinfo is None:
                # Make naive to match cryptography.
                now = now.replace(tzinfo=None)
            if this_update <= now < next_update:
                return value

            self._data.pop(cache_key, None)
            raise KeyError(cache_key)



# Copyright 2009-2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some tools for running tests based on MongoDB server version."""
from __future__ import annotations

from pymongo.common import Version as BaseVersion


class Version(BaseVersion):
    @classmethod
    def from_client(cls, client):
        info = client.server_info()
        if "versionArray" in info:
            return cls.from_version_array(info["versionArray"])
        return cls.from_string(info["version"])

    @classmethod
    async def async_from_client(cls, client):
        info = await client.server_info()
        if "versionArray" in info:
            return cls.from_version_array(info["versionArray"])
        return cls.from_string(info["version"])



# Copyright 2021-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the Load Balancer unified spec tests."""
from __future__ import annotations

import asyncio
import gc
import os
import pathlib
import sys
import threading
from asyncio import Event
from test.asynchronous.helpers import ConcurrentRunner, ExceptionCatchingTask
from test.asynchronous.utils import async_get_pool

import pytest

sys.path[0:0] = [""]

from test.asynchronous import AsyncIntegrationTest, async_client_context, unittest
from test.asynchronous.unified_format import generate_test_classes, get_test_path
from test.utils_shared import (
    async_wait_until,
    create_async_event,
)

from pymongo.asynchronous.helpers import anext

_IS_SYNC = False

pytestmark = pytest.mark.load_balancer

# Generate unified tests.
globals().update(generate_test_classes(get_test_path("load_balancer"), module=__name__))


class TestLB(AsyncIntegrationTest):
    RUN_ON_LOAD_BALANCER = True

    async def test_connections_are_only_returned_once(self):
        if "PyPy" in sys.version:
            # Tracked in PYTHON-3011
            self.skipTest("Test is flaky on PyPy")
        pool = await async_get_pool(self.client)
        n_conns = len(pool.conns)
        await self.db.test.find_one({})
        self.assertEqual(len(pool.conns), n_conns)
        await (await self.db.test.aggregate([{"$limit": 1}])).to_list()
        self.assertEqual(len(pool.conns), n_conns)

    @async_client_context.require_load_balancer
    async def test_unpin_committed_transaction(self):
        client = await self.async_rs_client()
        pool = await async_get_pool(client)
        coll = client[self.db.name].test
        async with client.start_session() as session:
            async with await session.start_transaction():
                self.assertEqual(pool.active_sockets, 0)
                await coll.insert_one({}, session=session)
                self.assertEqual(pool.active_sockets, 1)  # Pinned.
            self.assertEqual(pool.active_sockets, 1)  # Still pinned.
        self.assertEqual(pool.active_sockets, 0)  # Unpinned.

    @async_client_context.require_failCommand_fail_point
    async def test_cursor_gc(self):
        async def create_resource(coll):
            cursor = coll.find({}, batch_size=3)
            await anext(cursor)
            return cursor

        await self._test_no_gc_deadlock(create_resource)

    @async_client_context.require_failCommand_fail_point
    async def test_command_cursor_gc(self):
        async def create_resource(coll):
            cursor = await coll.aggregate([], batchSize=3)
            await anext(cursor)
            return cursor

        await self._test_no_gc_deadlock(create_resource)

    async def _test_no_gc_deadlock(self, create_resource):
        client = await self.async_rs_client()
        pool = await async_get_pool(client)
        coll = client[self.db.name].test
        await coll.insert_many([{} for _ in range(10)])
        self.assertEqual(pool.active_sockets, 0)
        # Cause the initial find attempt to fail to induce a reference cycle.
        args = {
            "mode": {"times": 1},
            "data": {
                "failCommands": ["find", "aggregate"],
                "closeConnection": True,
            },
        }
        async with self.fail_point(args):
            resource = await create_resource(coll)
            if async_client_context.load_balancer:
                self.assertEqual(pool.active_sockets, 1)  # Pinned.

        task = PoolLocker(pool)
        await task.start()
        self.assertTrue(await task.wait(task.locked, 5), "timed out")
        # Garbage collect the resource while the pool is locked to ensure we
        # don't deadlock.
        del resource
        # On PyPy it can take a few rounds to collect the cursor.
        for _ in range(3):
            gc.collect()
        task.unlock.set()
        await task.join(5)
        self.assertFalse(task.is_alive())
        self.assertIsNone(task.exc)

        await async_wait_until(lambda: pool.active_sockets == 0, "return socket")
        # Run another operation to ensure the socket still works.
        await coll.delete_many({})

    @async_client_context.require_transactions
    async def test_session_gc(self):
        client = await self.async_rs_client()
        pool = await async_get_pool(client)
        session = client.start_session()
        await session.start_transaction()
        await client.test_session_gc.test.find_one({}, session=session)
        # Cleanup the transaction left open on the server
        self.addAsyncCleanup(self.client.admin.command, "killSessions", [session.session_id])
        if async_client_context.load_balancer:
            self.assertEqual(pool.active_sockets, 1)  # Pinned.

        task = PoolLocker(pool)
        await task.start()
        self.assertTrue(await task.wait(task.locked, 5), "timed out")
        # Garbage collect the session while the pool is locked to ensure we
        # don't deadlock.
        del session
        # On PyPy it can take a few rounds to collect the session.
        for _ in range(3):
            gc.collect()
        task.unlock.set()
        await task.join(5)
        self.assertFalse(task.is_alive())
        self.assertIsNone(task.exc)

        await async_wait_until(lambda: pool.active_sockets == 0, "return socket")
        # Run another operation to ensure the socket still works.
        await client[self.db.name].test.delete_many({})


class PoolLocker(ExceptionCatchingTask):
    def __init__(self, pool):
        super().__init__(target=self.lock_pool)
        self.pool = pool
        self.daemon = True
        self.locked = create_async_event()
        self.unlock = create_async_event()

    async def lock_pool(self):
        async with self.pool.lock:
            self.locked.set()
            # Wait for the unlock flag.
            unlock_pool = await self.wait(self.unlock, 10)
            if not unlock_pool:
                raise Exception("timed out waiting for unlock signal: deadlock?")

    async def wait(self, event: Event, timeout: int):
        if _IS_SYNC:
            return event.wait(timeout)  # type: ignore[call-arg]
        else:
            try:
                await asyncio.wait_for(event.wait(), timeout=timeout)
            except asyncio.TimeoutError:
                return False
            return True


if __name__ == "__main__":
    unittest.main()



# Copyright 2009-2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fail if the C extension module doesn't exist.

Only really intended to be used by internal build scripts.
"""
from __future__ import annotations

import logging
import sys

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(levelname)-8s %(message)s")

sys.path[0:0] = [""]

import bson  # noqa: E402
import pymongo  # noqa: E402


def main() -> None:
    if not pymongo.has_c() or not bson.has_c():
        try:
            from pymongo import _cmessage  # type:ignore[attr-defined] # noqa: F401
        except Exception as e:
            LOGGER.exception(e)
        try:
            from bson import _cbson  # type:ignore[attr-defined] # noqa: F401
        except Exception as e:
            LOGGER.exception(e)
        sys.exit("could not load C extensions")


if __name__ == "__main__":
    main()



# Copyright 2012-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared utilities for testing pymongo"""
from __future__ import annotations

import asyncio
import contextlib
import copy
import functools
import random
import re
import shutil
import sys
import threading
import unittest
import warnings
from collections import abc, defaultdict
from functools import partial
from inspect import iscoroutinefunction
from test import client_context
from test.asynchronous.utils import async_wait_until
from test.utils import wait_until
from typing import List

from bson.objectid import ObjectId
from pymongo import monitoring, operations, read_preferences
from pymongo.cursor_shared import CursorType
from pymongo.errors import OperationFailure
from pymongo.helpers_shared import _SENSITIVE_COMMANDS
from pymongo.lock import _async_create_lock, _create_lock
from pymongo.monitoring import (
    ConnectionCheckedInEvent,
    ConnectionCheckedOutEvent,
    ConnectionCheckOutFailedEvent,
    ConnectionCheckOutStartedEvent,
    ConnectionClosedEvent,
    ConnectionCreatedEvent,
    ConnectionReadyEvent,
    PoolClearedEvent,
    PoolClosedEvent,
    PoolCreatedEvent,
    PoolReadyEvent,
)
from pymongo.read_concern import ReadConcern
from pymongo.server_type import SERVER_TYPE
from pymongo.synchronous.collection import ReturnDocument
from pymongo.synchronous.pool import _CancellationContext, _PoolGeneration
from pymongo.write_concern import WriteConcern

IMPOSSIBLE_WRITE_CONCERN = WriteConcern(w=50)


class BaseListener:
    def __init__(self):
        self.events = []

    def reset(self):
        self.events = []

    def add_event(self, event):
        self.events.append(event)

    def event_count(self, event_type):
        return len(self.events_by_type(event_type))

    def events_by_type(self, event_type):
        """Return the matching events by event class.

        event_type can be a single class or a tuple of classes.
        """
        return self.matching(lambda e: isinstance(e, event_type))

    def matching(self, matcher):
        """Return the matching events."""
        return [event for event in self.events[:] if matcher(event)]

    def wait_for_event(self, event, count):
        """Wait for a number of events to be published, or fail."""
        wait_until(lambda: self.event_count(event) >= count, f"find {count} {event} event(s)")

    async def async_wait_for_event(self, event, count):
        """Wait for a number of events to be published, or fail."""
        await async_wait_until(
            lambda: self.event_count(event) >= count, f"find {count} {event} event(s)"
        )


class CMAPListener(BaseListener, monitoring.ConnectionPoolListener):
    def connection_created(self, event):
        assert isinstance(event, ConnectionCreatedEvent)
        self.add_event(event)

    def connection_ready(self, event):
        assert isinstance(event, ConnectionReadyEvent)
        self.add_event(event)

    def connection_closed(self, event):
        assert isinstance(event, ConnectionClosedEvent)
        self.add_event(event)

    def connection_check_out_started(self, event):
        assert isinstance(event, ConnectionCheckOutStartedEvent)
        self.add_event(event)

    def connection_check_out_failed(self, event):
        assert isinstance(event, ConnectionCheckOutFailedEvent)
        self.add_event(event)

    def connection_checked_out(self, event):
        assert isinstance(event, ConnectionCheckedOutEvent)
        self.add_event(event)

    def connection_checked_in(self, event):
        assert isinstance(event, ConnectionCheckedInEvent)
        self.add_event(event)

    def pool_created(self, event):
        assert isinstance(event, PoolCreatedEvent)
        self.add_event(event)

    def pool_ready(self, event):
        assert isinstance(event, PoolReadyEvent)
        self.add_event(event)

    def pool_cleared(self, event):
        assert isinstance(event, PoolClearedEvent)
        self.add_event(event)

    def pool_closed(self, event):
        assert isinstance(event, PoolClosedEvent)
        self.add_event(event)


class EventListener(BaseListener, monitoring.CommandListener):
    def __init__(self):
        super().__init__()
        self.results = defaultdict(list)

    @property
    def started_events(self) -> List[monitoring.CommandStartedEvent]:
        return self.results["started"]

    @property
    def succeeded_events(self) -> List[monitoring.CommandSucceededEvent]:
        return self.results["succeeded"]

    @property
    def failed_events(self) -> List[monitoring.CommandFailedEvent]:
        return self.results["failed"]

    def started(self, event: monitoring.CommandStartedEvent) -> None:
        self.started_events.append(event)
        self.add_event(event)

    def succeeded(self, event: monitoring.CommandSucceededEvent) -> None:
        self.succeeded_events.append(event)
        self.add_event(event)

    def failed(self, event: monitoring.CommandFailedEvent) -> None:
        self.failed_events.append(event)
        self.add_event(event)

    def started_command_names(self) -> List[str]:
        """Return list of command names started."""
        return [event.command_name for event in self.started_events]

    def reset(self) -> None:
        """Reset the state of this listener."""
        self.results.clear()
        super().reset()


class TopologyEventListener(monitoring.TopologyListener):
    def __init__(self):
        self.results = defaultdict(list)

    def closed(self, event):
        self.results["closed"].append(event)

    def description_changed(self, event):
        self.results["description_changed"].append(event)

    def opened(self, event):
        self.results["opened"].append(event)

    def reset(self):
        """Reset the state of this listener."""
        self.results.clear()


class AllowListEventListener(EventListener):
    def __init__(self, *commands):
        self.commands = set(commands)
        super().__init__()

    def started(self, event):
        if event.command_name in self.commands:
            super().started(event)

    def succeeded(self, event):
        if event.command_name in self.commands:
            super().succeeded(event)

    def failed(self, event):
        if event.command_name in self.commands:
            super().failed(event)


class OvertCommandListener(EventListener):
    """A CommandListener that ignores sensitive commands."""

    ignore_list_collections = False

    def started(self, event):
        if event.command_name.lower() not in _SENSITIVE_COMMANDS:
            super().started(event)

    def succeeded(self, event):
        if event.command_name.lower() not in _SENSITIVE_COMMANDS:
            super().succeeded(event)

    def failed(self, event):
        if event.command_name.lower() not in _SENSITIVE_COMMANDS:
            super().failed(event)


class _ServerEventListener:
    """Listens to all events."""

    def __init__(self):
        self.results = []

    def opened(self, event):
        self.results.append(event)

    def description_changed(self, event):
        self.results.append(event)

    def closed(self, event):
        self.results.append(event)

    def matching(self, matcher):
        """Return the matching events."""
        results = self.results[:]
        return [event for event in results if matcher(event)]

    def reset(self):
        self.results = []


class ServerEventListener(_ServerEventListener, monitoring.ServerListener):
    """Listens to Server events."""


class ServerAndTopologyEventListener(  # type: ignore[misc]
    ServerEventListener, monitoring.TopologyListener
):
    """Listens to Server and Topology events."""


class HeartbeatEventListener(BaseListener, monitoring.ServerHeartbeatListener):
    """Listens to only server heartbeat events."""

    def started(self, event):
        self.add_event(event)

    def succeeded(self, event):
        self.add_event(event)

    def failed(self, event):
        self.add_event(event)


class HeartbeatEventsListListener(HeartbeatEventListener):
    """Listens to only server heartbeat events and publishes them to a provided list."""

    def __init__(self, events):
        super().__init__()
        self.event_list = events

    def started(self, event):
        self.add_event(event)
        self.event_list.append("serverHeartbeatStartedEvent")

    def succeeded(self, event):
        self.add_event(event)
        self.event_list.append("serverHeartbeatSucceededEvent")

    def failed(self, event):
        self.add_event(event)
        self.event_list.append("serverHeartbeatFailedEvent")


class ScenarioDict(dict):
    """Dict that returns {} for any unknown key, recursively."""

    def __init__(self, data):
        def convert(v):
            if isinstance(v, abc.Mapping):
                return ScenarioDict(v)
            if isinstance(v, (str, bytes)):
                return v
            if isinstance(v, abc.Sequence):
                return [convert(item) for item in v]
            return v

        dict.__init__(self, [(k, convert(v)) for k, v in data.items()])

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            # Unlike a defaultdict, don't set the key, just return a dict.
            return ScenarioDict({})


class CompareType:
    """Class that compares equal to any object of the given type(s)."""

    def __init__(self, types):
        self.types = types

    def __eq__(self, other):
        return isinstance(other, self.types)


class FunctionCallRecorder:
    """Utility class to wrap a callable and record its invocations."""

    def __init__(self, function):
        self._function = function
        self._call_list = []

    def __call__(self, *args, **kwargs):
        self._call_list.append((args, kwargs))
        if iscoroutinefunction(self._function):
            return self._function(*args, **kwargs)
        else:
            return self._function(*args, **kwargs)

    def reset(self):
        """Wipes the call list."""
        self._call_list = []

    def call_list(self):
        """Returns a copy of the call list."""
        return self._call_list[:]

    @property
    def call_count(self):
        """Returns the number of times the function has been called."""
        return len(self._call_list)


def one(s):
    """Get one element of a set"""
    return next(iter(s))


def oid_generated_on_process(oid):
    """Makes a determination as to whether the given ObjectId was generated
    by the current process, based on the 5-byte random number in the ObjectId.
    """
    return ObjectId._random() == oid.binary[4:9]


def delay(sec):
    """Along with a ``$where`` operator, this triggers an arbitrarily long-running
    operation on the server.

    This can be useful in time-sensitive tests (e.g., timeouts, signals).
    Note that you must have at least one document in the collection or the
    server may decide not to sleep at all.

    Example
    -------

    .. code-block:: python

        db.coll.insert_one({"x": 1})
        db.test.find_one({"x": 1})
        # {'x': 1, '_id': ObjectId('54f4e12bfba5220aa4d6dee8')}

        # The following will wait 2.5 seconds before returning.
        db.test.find_one({"$where": delay(2.5)})
        # {'x': 1, '_id': ObjectId('54f4e12bfba5220aa4d6dee8')}

    Using ``delay`` to provoke a KeyboardInterrupt
    ----------------------------------------------

    .. code-block:: python

        import signal

        # Raise KeyboardInterrupt in 1 second
        def sigalarm(num, frame):
            raise KeyboardInterrupt


        signal.signal(signal.SIGALRM, sigalarm)
        signal.alarm(1)

        raised = False
        try:
            clxn.find_one({"$where": delay(1.5)})
        except KeyboardInterrupt:
            raised = True

        assert raised
    """
    return "function() { sleep(%f * 1000); return true; }" % sec


def camel_to_snake(camel):
    # Regex to convert CamelCase to snake_case.
    snake = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", camel)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", snake).lower()


def camel_to_upper_camel(camel):
    return camel[0].upper() + camel[1:]


def camel_to_snake_args(arguments):
    for arg_name in list(arguments):
        c2s = camel_to_snake(arg_name)
        arguments[c2s] = arguments.pop(arg_name)
    return arguments


def snake_to_camel(snake):
    # Regex to convert snake_case to lowerCamelCase.
    return re.sub(r"_([a-z])", lambda m: m.group(1).upper(), snake)


def parse_collection_options(opts):
    if "readPreference" in opts:
        opts["read_preference"] = parse_read_preference(opts.pop("readPreference"))

    if "writeConcern" in opts:
        opts["write_concern"] = WriteConcern(**dict(opts.pop("writeConcern")))

    if "readConcern" in opts:
        opts["read_concern"] = ReadConcern(**dict(opts.pop("readConcern")))

    if "timeoutMS" in opts:
        opts["timeout"] = int(opts.pop("timeoutMS")) / 1000.0
    return opts


@contextlib.contextmanager
def _ignore_deprecations():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        yield


def ignore_deprecations(wrapped=None):
    """A context manager or a decorator."""
    if wrapped:
        if iscoroutinefunction(wrapped):

            @functools.wraps(wrapped)
            async def wrapper(*args, **kwargs):
                with _ignore_deprecations():
                    return await wrapped(*args, **kwargs)
        else:

            @functools.wraps(wrapped)
            def wrapper(*args, **kwargs):
                with _ignore_deprecations():
                    return wrapped(*args, **kwargs)

        return wrapper

    else:
        return _ignore_deprecations()


class DeprecationFilter:
    def __init__(self, action="ignore"):
        """Start filtering deprecations."""
        self.warn_context = warnings.catch_warnings()
        self.warn_context.__enter__()
        warnings.simplefilter(action, DeprecationWarning)

    def stop(self):
        """Stop filtering deprecations."""
        self.warn_context.__exit__()  # type: ignore
        self.warn_context = None  # type: ignore


# Constants for run_threads and lazy_client_trial.
NTRIALS = 5
NTHREADS = 10


def run_threads(collection, target):
    """Run a target function in many threads.

    target is a function taking a Collection and an integer.
    """
    threads = []
    for i in range(NTHREADS):
        bound_target = partial(target, collection, i)
        threads.append(threading.Thread(target=bound_target))

    for t in threads:
        t.start()

    for t in threads:
        t.join(60)
        assert not t.is_alive()


@contextlib.contextmanager
def frequent_thread_switches():
    """Make concurrency bugs more likely to manifest."""
    interval = sys.getswitchinterval()
    sys.setswitchinterval(1e-6)

    try:
        yield
    finally:
        sys.setswitchinterval(interval)


def lazy_client_trial(reset, target, test, get_client):
    """Test concurrent operations on a lazily-connecting client.

    `reset` takes a collection and resets it for the next trial.

    `target` takes a lazily-connecting collection and an index from
    0 to NTHREADS, and performs some operation, e.g. an insert.

    `test` takes the lazily-connecting collection and asserts a
    post-condition to prove `target` succeeded.
    """
    collection = client_context.client.pymongo_test.test

    with frequent_thread_switches():
        for _i in range(NTRIALS):
            reset(collection)
            lazy_client = get_client()
            lazy_collection = lazy_client.pymongo_test.test
            run_threads(lazy_collection, target)
            test(lazy_collection)


def gevent_monkey_patched():
    """Check if gevent's monkey patching is active."""
    try:
        import socket

        import gevent.socket  # type:ignore[import]

        return socket.socket is gevent.socket.socket
    except ImportError:
        return False


def is_greenthread_patched():
    return gevent_monkey_patched()


def parse_read_preference(pref):
    # Make first letter lowercase to match read_pref's modes.
    mode_string = pref.get("mode", "primary")
    mode_string = mode_string[:1].lower() + mode_string[1:]
    mode = read_preferences.read_pref_mode_from_name(mode_string)
    max_staleness = pref.get("maxStalenessSeconds", -1)
    tag_sets = pref.get("tagSets") or pref.get("tag_sets")
    return read_preferences.make_read_preference(
        mode, tag_sets=tag_sets, max_staleness=max_staleness
    )


def server_name_to_type(name):
    """Convert a ServerType name to the corresponding value. For SDAM tests."""
    # Special case, some tests in the spec include the PossiblePrimary
    # type, but only single-threaded drivers need that type. We call
    # possible primaries Unknown.
    if name == "PossiblePrimary":
        return SERVER_TYPE.Unknown
    return getattr(SERVER_TYPE, name)


def cat_files(dest, *sources):
    """Cat multiple files into dest."""
    with open(dest, "wb") as fdst:
        for src in sources:
            with open(src, "rb") as fsrc:
                shutil.copyfileobj(fsrc, fdst)


@contextlib.contextmanager
def assertion_context(msg):
    """A context manager that adds info to an assertion failure."""
    try:
        yield
    except AssertionError as exc:
        raise AssertionError(f"{msg}: {exc}")


def parse_spec_options(opts):
    if "readPreference" in opts:
        opts["read_preference"] = parse_read_preference(opts.pop("readPreference"))

    if "writeConcern" in opts:
        w_opts = opts.pop("writeConcern")
        if "journal" in w_opts:
            w_opts["j"] = w_opts.pop("journal")
        if "wtimeoutMS" in w_opts:
            w_opts["wtimeout"] = w_opts.pop("wtimeoutMS")
        opts["write_concern"] = WriteConcern(**dict(w_opts))

    if "readConcern" in opts:
        opts["read_concern"] = ReadConcern(**dict(opts.pop("readConcern")))

    if "timeoutMS" in opts:
        assert isinstance(opts["timeoutMS"], int)
        opts["timeout"] = int(opts.pop("timeoutMS")) / 1000.0

    if "maxTimeMS" in opts:
        opts["max_time_ms"] = opts.pop("maxTimeMS")

    if "maxCommitTimeMS" in opts:
        opts["max_commit_time_ms"] = opts.pop("maxCommitTimeMS")

    return dict(opts)


def prepare_spec_arguments(spec, arguments, opname, entity_map, with_txn_callback):
    for arg_name in list(arguments):
        c2s = camel_to_snake(arg_name)
        # Named "key" instead not fieldName.
        if arg_name == "fieldName":
            arguments["key"] = arguments.pop(arg_name)
        # Aggregate uses "batchSize", while find uses batch_size.
        elif (arg_name == "batchSize" or arg_name == "allowDiskUse") and opname == "aggregate":
            continue
        elif arg_name == "bypassDocumentValidation" and (
            opname == "aggregate" or "find_one_and" in opname
        ):
            continue
        elif arg_name == "timeoutMode":
            raise unittest.SkipTest("PyMongo does not support timeoutMode")
        # Requires boolean returnDocument.
        elif arg_name == "returnDocument":
            arguments[c2s] = getattr(ReturnDocument, arguments.pop(arg_name).upper())
        elif "bulk_write" in opname and (c2s == "requests" or c2s == "models"):
            # Parse each request into a bulk write model.
            requests = []
            for request in arguments[c2s]:
                if "name" in request:
                    # CRUD v2 format
                    bulk_model = camel_to_upper_camel(request["name"])
                    bulk_class = getattr(operations, bulk_model)
                    bulk_arguments = camel_to_snake_args(request["arguments"])
                else:
                    # Unified test format
                    bulk_model, spec = next(iter(request.items()))
                    bulk_class = getattr(operations, camel_to_upper_camel(bulk_model))
                    bulk_arguments = camel_to_snake_args(spec)
                requests.append(bulk_class(**dict(bulk_arguments)))
            arguments[c2s] = requests
        elif arg_name == "session":
            arguments["session"] = entity_map[arguments["session"]]
        elif opname == "open_download_stream" and arg_name == "id":
            arguments["file_id"] = arguments.pop(arg_name)
        elif opname not in ("find", "find_one") and c2s == "max_time_ms":
            # find is the only method that accepts snake_case max_time_ms.
            # All other methods take kwargs which must use the server's
            # camelCase maxTimeMS. See PYTHON-1855.
            arguments["maxTimeMS"] = arguments.pop("max_time_ms")
        elif opname == "with_transaction" and arg_name == "callback":
            if "operations" in arguments[arg_name]:
                # CRUD v2 format
                callback_ops = arguments[arg_name]["operations"]
            else:
                # Unified test format
                callback_ops = arguments[arg_name]
            arguments["callback"] = lambda _: with_txn_callback(copy.deepcopy(callback_ops))
        elif opname == "drop_collection" and arg_name == "collection":
            arguments["name_or_collection"] = arguments.pop(arg_name)
        elif opname == "create_collection":
            if arg_name == "collection":
                arguments["name"] = arguments.pop(arg_name)
            arguments["check_exists"] = False
            # Any other arguments to create_collection are passed through
            # **kwargs.
        elif opname == "create_index" and arg_name == "keys":
            arguments["keys"] = list(arguments.pop(arg_name).items())
        elif opname == "drop_index" and arg_name == "name":
            arguments["index_or_name"] = arguments.pop(arg_name)
        elif opname == "rename" and arg_name == "to":
            arguments["new_name"] = arguments.pop(arg_name)
        elif opname == "rename" and arg_name == "dropTarget":
            arguments["dropTarget"] = arguments.pop(arg_name)
        elif arg_name == "cursorType":
            cursor_type = arguments.pop(arg_name)
            if cursor_type == "tailable":
                arguments["cursor_type"] = CursorType.TAILABLE
            elif cursor_type == "tailableAwait":
                arguments["cursor_type"] = CursorType.TAILABLE
            else:
                raise AssertionError(f"Unsupported cursorType: {cursor_type}")
        else:
            arguments[c2s] = arguments.pop(arg_name)


def create_async_event():
    return asyncio.Event()


def create_event():
    return threading.Event()


def async_create_barrier(n_tasks: int):
    return asyncio.Barrier(n_tasks)


def create_barrier(n_tasks: int, timeout: float | None = None):
    return threading.Barrier(n_tasks, timeout=timeout)


async def async_barrier_wait(barrier, timeout: float | None = None):
    await asyncio.wait_for(barrier.wait(), timeout=timeout)


def barrier_wait(barrier, timeout: float | None = None):
    barrier.wait(timeout=timeout)



#
# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the grid_file module."""
from __future__ import annotations

import datetime
import io
import sys
import zipfile
from io import BytesIO
from test.asynchronous import (
    AsyncIntegrationTest,
    AsyncUnitTest,
    async_client_context,
    qcheck,
    unittest,
)

from pymongo.asynchronous.database import AsyncDatabase

sys.path[0:0] = [""]

from test.utils_shared import OvertCommandListener

from bson.objectid import ObjectId
from gridfs.asynchronous.grid_file import (
    _SEEK_CUR,
    _SEEK_END,
    DEFAULT_CHUNK_SIZE,
    AsyncGridFS,
    AsyncGridIn,
    AsyncGridOut,
    AsyncGridOutCursor,
)
from gridfs.errors import NoFile
from pymongo import AsyncMongoClient
from pymongo.asynchronous.helpers import aiter, anext
from pymongo.errors import ConfigurationError, ServerSelectionTimeoutError
from pymongo.message import _CursorAddress

_IS_SYNC = False


class AsyncTestGridFileNoConnect(AsyncUnitTest):
    """Test GridFile features on a client that does not connect."""

    db: AsyncDatabase

    @classmethod
    def setUpClass(cls):
        cls.db = AsyncMongoClient(connect=False).pymongo_test

    def test_grid_in_custom_opts(self):
        self.assertRaises(TypeError, AsyncGridIn, "foo")

        a = AsyncGridIn(
            self.db.fs,
            _id=5,
            filename="my_file",
            contentType="text/html",
            chunkSize=1000,
            aliases=["foo"],
            metadata={"foo": 1, "bar": 2},
            bar=3,
            baz="hello",
        )

        self.assertEqual(5, a._id)
        self.assertEqual("my_file", a.filename)
        self.assertEqual("my_file", a.name)
        self.assertEqual("text/html", a.content_type)
        self.assertEqual(1000, a.chunk_size)
        self.assertEqual(["foo"], a.aliases)
        self.assertEqual({"foo": 1, "bar": 2}, a.metadata)
        self.assertEqual(3, a.bar)
        self.assertEqual("hello", a.baz)
        self.assertRaises(AttributeError, getattr, a, "mike")

        b = AsyncGridIn(self.db.fs, content_type="text/html", chunk_size=1000, baz=100)
        self.assertEqual("text/html", b.content_type)
        self.assertEqual(1000, b.chunk_size)
        self.assertEqual(100, b.baz)


class AsyncTestGridFile(AsyncIntegrationTest):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        await self.cleanup_colls(self.db.fs.files, self.db.fs.chunks)

    async def test_basic(self):
        f = AsyncGridIn(self.db.fs, filename="test")
        await f.write(b"hello world")
        await f.close()
        self.assertEqual(1, await self.db.fs.files.count_documents({}))
        self.assertEqual(1, await self.db.fs.chunks.count_documents({}))

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"hello world", await g.read())

        # make sure it's still there...
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"hello world", await g.read())

        f = AsyncGridIn(self.db.fs, filename="test")
        await f.close()
        self.assertEqual(2, await self.db.fs.files.count_documents({}))
        self.assertEqual(1, await self.db.fs.chunks.count_documents({}))

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"", await g.read())

        # test that reading 0 returns proper type
        self.assertEqual(b"", await g.read(0))

    async def test_md5(self):
        f = AsyncGridIn(self.db.fs)
        await f.write(b"hello world\n")
        await f.close()
        self.assertEqual(None, f.md5)

    async def test_alternate_collection(self):
        await self.db.alt.files.delete_many({})
        await self.db.alt.chunks.delete_many({})

        f = AsyncGridIn(self.db.alt)
        await f.write(b"hello world")
        await f.close()

        self.assertEqual(1, await self.db.alt.files.count_documents({}))
        self.assertEqual(1, await self.db.alt.chunks.count_documents({}))

        g = AsyncGridOut(self.db.alt, f._id)
        self.assertEqual(b"hello world", await g.read())

    async def test_grid_in_default_opts(self):
        self.assertRaises(TypeError, AsyncGridIn, "foo")

        a = AsyncGridIn(self.db.fs)

        self.assertIsInstance(a._id, ObjectId)
        self.assertRaises(AttributeError, setattr, a, "_id", 5)

        self.assertEqual(None, a.filename)
        self.assertEqual(None, a.name)
        a.filename = "my_file"
        self.assertEqual("my_file", a.filename)
        self.assertEqual("my_file", a.name)

        self.assertEqual(None, a.content_type)
        a.content_type = "text/html"

        self.assertEqual("text/html", a.content_type)

        self.assertRaises(AttributeError, getattr, a, "length")
        self.assertRaises(AttributeError, setattr, a, "length", 5)

        self.assertEqual(255 * 1024, a.chunk_size)
        self.assertRaises(AttributeError, setattr, a, "chunk_size", 5)

        self.assertRaises(AttributeError, getattr, a, "upload_date")
        self.assertRaises(AttributeError, setattr, a, "upload_date", 5)

        self.assertRaises(AttributeError, getattr, a, "aliases")
        a.aliases = ["foo"]

        self.assertEqual(["foo"], a.aliases)

        self.assertRaises(AttributeError, getattr, a, "metadata")
        a.metadata = {"foo": 1}

        self.assertEqual({"foo": 1}, a.metadata)

        self.assertRaises(AttributeError, setattr, a, "md5", 5)

        await a.close()

        if _IS_SYNC:
            a.forty_two = 42
        else:
            self.assertRaises(AttributeError, setattr, a, "forty_two", 42)
            await a.set("forty_two", 42)

        self.assertEqual(42, a.forty_two)

        self.assertIsInstance(a._id, ObjectId)
        self.assertRaises(AttributeError, setattr, a, "_id", 5)

        self.assertEqual("my_file", a.filename)
        self.assertEqual("my_file", a.name)

        self.assertEqual("text/html", a.content_type)

        self.assertEqual(0, a.length)
        self.assertRaises(AttributeError, setattr, a, "length", 5)

        self.assertEqual(255 * 1024, a.chunk_size)
        self.assertRaises(AttributeError, setattr, a, "chunk_size", 5)

        self.assertIsInstance(a.upload_date, datetime.datetime)
        self.assertRaises(AttributeError, setattr, a, "upload_date", 5)

        self.assertEqual(["foo"], a.aliases)

        self.assertEqual({"foo": 1}, a.metadata)

        self.assertEqual(None, a.md5)
        self.assertRaises(AttributeError, setattr, a, "md5", 5)

        # Make sure custom attributes that were set both before and after
        # a.close() are reflected in b. PYTHON-411.
        b = await AsyncGridFS(self.db).get_last_version(filename=a.filename)
        self.assertEqual(a.metadata, b.metadata)
        self.assertEqual(a.aliases, b.aliases)
        self.assertEqual(a.forty_two, b.forty_two)

    async def test_grid_out_default_opts(self):
        self.assertRaises(TypeError, AsyncGridOut, "foo")

        gout = AsyncGridOut(self.db.fs, 5)
        with self.assertRaises(NoFile):
            if not _IS_SYNC:
                await gout.open()
            gout.name

        a = AsyncGridIn(self.db.fs)
        await a.close()

        b = AsyncGridOut(self.db.fs, a._id)
        if not _IS_SYNC:
            await b.open()

        self.assertEqual(a._id, b._id)
        self.assertEqual(0, b.length)
        self.assertEqual(None, b.content_type)
        self.assertEqual(None, b.name)
        self.assertEqual(None, b.filename)
        self.assertEqual(255 * 1024, b.chunk_size)
        self.assertIsInstance(b.upload_date, datetime.datetime)
        self.assertEqual(None, b.aliases)
        self.assertEqual(None, b.metadata)
        self.assertEqual(None, b.md5)

        for attr in [
            "_id",
            "name",
            "content_type",
            "length",
            "chunk_size",
            "upload_date",
            "aliases",
            "metadata",
            "md5",
        ]:
            self.assertRaises(AttributeError, setattr, b, attr, 5)

    async def test_grid_out_cursor_options(self):
        self.assertRaises(
            TypeError, AsyncGridOutCursor.__init__, self.db.fs, {}, projection={"filename": 1}
        )

        cursor = AsyncGridOutCursor(self.db.fs, {})
        cursor_clone = cursor.clone()

        cursor_dict = cursor.__dict__.copy()
        cursor_dict.pop("_session")
        cursor_clone_dict = cursor_clone.__dict__.copy()
        cursor_clone_dict.pop("_session")
        self.assertDictEqual(cursor_dict, cursor_clone_dict)

        self.assertRaises(NotImplementedError, cursor.add_option, 0)
        self.assertRaises(NotImplementedError, cursor.remove_option, 0)

    async def test_grid_out_custom_opts(self):
        one = AsyncGridIn(
            self.db.fs,
            _id=5,
            filename="my_file",
            contentType="text/html",
            chunkSize=1000,
            aliases=["foo"],
            metadata={"foo": 1, "bar": 2},
            bar=3,
            baz="hello",
        )
        await one.write(b"hello world")
        await one.close()

        two = AsyncGridOut(self.db.fs, 5)

        if not _IS_SYNC:
            await two.open()

        self.assertEqual("my_file", two.name)
        self.assertEqual("my_file", two.filename)
        self.assertEqual(5, two._id)
        self.assertEqual(11, two.length)
        self.assertEqual("text/html", two.content_type)
        self.assertEqual(1000, two.chunk_size)
        self.assertIsInstance(two.upload_date, datetime.datetime)
        self.assertEqual(["foo"], two.aliases)
        self.assertEqual({"foo": 1, "bar": 2}, two.metadata)
        self.assertEqual(3, two.bar)
        self.assertEqual(None, two.md5)

        for attr in [
            "_id",
            "name",
            "content_type",
            "length",
            "chunk_size",
            "upload_date",
            "aliases",
            "metadata",
            "md5",
        ]:
            self.assertRaises(AttributeError, setattr, two, attr, 5)

    async def test_grid_out_file_document(self):
        one = AsyncGridIn(self.db.fs)
        await one.write(b"foo bar")
        await one.close()

        two = AsyncGridOut(self.db.fs, file_document=await self.db.fs.files.find_one())
        self.assertEqual(b"foo bar", await two.read())

        three = AsyncGridOut(self.db.fs, 5, file_document=await self.db.fs.files.find_one())
        self.assertEqual(b"foo bar", await three.read())

        four = AsyncGridOut(self.db.fs, file_document={})
        with self.assertRaises(NoFile):
            if not _IS_SYNC:
                await four.open()
            four.name

    async def test_write_file_like(self):
        one = AsyncGridIn(self.db.fs)
        await one.write(b"hello world")
        await one.close()

        two = AsyncGridOut(self.db.fs, one._id)

        three = AsyncGridIn(self.db.fs)
        await three.write(two)
        await three.close()

        four = AsyncGridOut(self.db.fs, three._id)
        self.assertEqual(b"hello world", await four.read())

        five = AsyncGridIn(self.db.fs, chunk_size=2)
        await five.write(b"hello")
        buffer = BytesIO(b" world")
        await five.write(buffer)
        await five.write(b" and mongodb")
        await five.close()
        self.assertEqual(
            b"hello world and mongodb", await AsyncGridOut(self.db.fs, five._id).read()
        )

    async def test_write_lines(self):
        a = AsyncGridIn(self.db.fs)
        await a.writelines([b"hello ", b"world"])
        await a.close()

        self.assertEqual(b"hello world", await AsyncGridOut(self.db.fs, a._id).read())

    async def test_close(self):
        f = AsyncGridIn(self.db.fs)
        await f.close()
        with self.assertRaises(ValueError):
            await f.write("test")
        await f.close()

    async def test_closed(self):
        f = AsyncGridIn(self.db.fs, chunkSize=5)
        await f.write(b"Hello world.\nHow are you?")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        if not _IS_SYNC:
            await g.open()
        self.assertFalse(g.closed)
        await g.read(1)
        self.assertFalse(g.closed)
        await g.read(100)
        self.assertFalse(g.closed)
        await g.close()
        self.assertTrue(g.closed)

    async def test_multi_chunk_file(self):
        random_string = b"a" * (DEFAULT_CHUNK_SIZE + 1000)

        f = AsyncGridIn(self.db.fs)
        await f.write(random_string)
        await f.close()

        self.assertEqual(1, await self.db.fs.files.count_documents({}))
        self.assertEqual(2, await self.db.fs.chunks.count_documents({}))

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(random_string, await g.read())

    async def test_small_chunks(self):
        self.files = 0
        self.chunks = 0

        async def helper(data):
            f = AsyncGridIn(self.db.fs, chunkSize=1)
            await f.write(data)
            await f.close()

            self.files += 1
            self.chunks += len(data)

            self.assertEqual(self.files, await self.db.fs.files.count_documents({}))
            self.assertEqual(self.chunks, await self.db.fs.chunks.count_documents({}))

            g = AsyncGridOut(self.db.fs, f._id)
            self.assertEqual(data, await g.read())

            g = AsyncGridOut(self.db.fs, f._id)
            self.assertEqual(data, await g.read(10) + await g.read(10))
            return True

        await qcheck.check_unittest(self, helper, qcheck.gen_string(qcheck.gen_range(0, 20)))

    async def test_seek(self):
        f = AsyncGridIn(self.db.fs, chunkSize=3)
        await f.write(b"hello world")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"hello world", await g.read())
        await g.seek(0)
        self.assertEqual(b"hello world", await g.read())
        await g.seek(1)
        self.assertEqual(b"ello world", await g.read())
        with self.assertRaises(IOError):
            await g.seek(-1)

        await g.seek(-3, _SEEK_END)
        self.assertEqual(b"rld", await g.read())
        await g.seek(0, _SEEK_END)
        self.assertEqual(b"", await g.read())
        with self.assertRaises(IOError):
            await g.seek(-100, _SEEK_END)

        await g.seek(3)
        await g.seek(3, _SEEK_CUR)
        self.assertEqual(b"world", await g.read())
        with self.assertRaises(IOError):
            await g.seek(-100, _SEEK_CUR)

    async def test_tell(self):
        f = AsyncGridIn(self.db.fs, chunkSize=3)
        await f.write(b"hello world")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(0, g.tell())
        await g.read(0)
        self.assertEqual(0, g.tell())
        await g.read(1)
        self.assertEqual(1, g.tell())
        await g.read(2)
        self.assertEqual(3, g.tell())
        await g.read()
        self.assertEqual(g.length, g.tell())

    async def test_multiple_reads(self):
        f = AsyncGridIn(self.db.fs, chunkSize=3)
        await f.write(b"hello world")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"he", await g.read(2))
        self.assertEqual(b"ll", await g.read(2))
        self.assertEqual(b"o ", await g.read(2))
        self.assertEqual(b"wo", await g.read(2))
        self.assertEqual(b"rl", await g.read(2))
        self.assertEqual(b"d", await g.read(2))
        self.assertEqual(b"", await g.read(2))

    async def test_readline(self):
        f = AsyncGridIn(self.db.fs, chunkSize=5)
        await f.write(
            b"""Hello world,
How are you?
Hope all is well.
Bye"""
        )
        await f.close()

        # Try read(), then readline().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"H", await g.read(1))
        self.assertEqual(b"ello world,\n", await g.readline())
        self.assertEqual(b"How a", await g.readline(5))
        self.assertEqual(b"", await g.readline(0))
        self.assertEqual(b"re you?\n", await g.readline())
        self.assertEqual(b"Hope all is well.\n", await g.readline(1000))
        self.assertEqual(b"Bye", await g.readline())
        self.assertEqual(b"", await g.readline())

        # Try readline() first, then read().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"He", await g.readline(2))
        self.assertEqual(b"l", await g.read(1))
        self.assertEqual(b"lo", await g.readline(2))
        self.assertEqual(b" world,\n", await g.readline())

        # Only readline().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"H", await g.readline(1))
        self.assertEqual(b"e", await g.readline(1))
        self.assertEqual(b"llo world,\n", await g.readline())

    async def test_readlines(self):
        f = AsyncGridIn(self.db.fs, chunkSize=5)
        await f.write(
            b"""Hello world,
How are you?
Hope all is well.
Bye"""
        )
        await f.close()

        # Try read(), then readlines().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"He", await g.read(2))
        self.assertEqual([b"llo world,\n", b"How are you?\n"], await g.readlines(11))
        self.assertEqual([b"Hope all is well.\n", b"Bye"], await g.readlines())
        self.assertEqual([], await g.readlines())

        # Try readline(), then readlines().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"Hello world,\n", await g.readline())
        self.assertEqual([b"How are you?\n", b"Hope all is well.\n"], await g.readlines(13))
        self.assertEqual(b"Bye", await g.readline())
        self.assertEqual([], await g.readlines())

        # Only readlines().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(
            [b"Hello world,\n", b"How are you?\n", b"Hope all is well.\n", b"Bye"],
            await g.readlines(),
        )

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(
            [b"Hello world,\n", b"How are you?\n", b"Hope all is well.\n", b"Bye"],
            await g.readlines(0),
        )

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual([b"Hello world,\n"], await g.readlines(1))
        self.assertEqual([b"How are you?\n"], await g.readlines(12))
        self.assertEqual([b"Hope all is well.\n", b"Bye"], await g.readlines(18))

        # Try readlines() first, then read().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual([b"Hello world,\n"], await g.readlines(1))
        self.assertEqual(b"H", await g.read(1))
        self.assertEqual([b"ow are you?\n", b"Hope all is well.\n"], await g.readlines(29))
        self.assertEqual([b"Bye"], await g.readlines(1))

        # Try readlines() first, then readline().
        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual([b"Hello world,\n"], await g.readlines(1))
        self.assertEqual(b"How are you?\n", await g.readline())
        self.assertEqual([b"Hope all is well.\n"], await g.readlines(17))
        self.assertEqual(b"Bye", await g.readline())

    async def test_iterator(self):
        f = AsyncGridIn(self.db.fs)
        await f.close()
        g = AsyncGridOut(self.db.fs, f._id)
        if _IS_SYNC:
            self.assertEqual([], list(g))
        else:
            self.assertEqual([], await g.to_list())

        f = AsyncGridIn(self.db.fs)
        await f.write(b"hello world\nhere are\nsome lines.")
        await f.close()
        g = AsyncGridOut(self.db.fs, f._id)
        if _IS_SYNC:
            self.assertEqual([b"hello world\n", b"here are\n", b"some lines."], list(g))
        else:
            self.assertEqual([b"hello world\n", b"here are\n", b"some lines."], await g.to_list())

        self.assertEqual(b"", await g.read(5))
        if _IS_SYNC:
            self.assertEqual([], list(g))
        else:
            self.assertEqual([], await g.to_list())

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"hello world\n", await anext(aiter(g)))
        self.assertEqual(b"here", await g.read(4))
        self.assertEqual(b" are\n", await anext(aiter(g)))
        self.assertEqual(b"some lines", await g.read(10))
        self.assertEqual(b".", await anext(aiter(g)))
        with self.assertRaises(StopAsyncIteration):
            await aiter(g).__anext__()

        f = AsyncGridIn(self.db.fs, chunk_size=2)
        await f.write(b"hello world")
        await f.close()
        g = AsyncGridOut(self.db.fs, f._id)
        if _IS_SYNC:
            self.assertEqual([b"hello world"], list(g))
        else:
            self.assertEqual([b"hello world"], await g.to_list())

    async def test_read_unaligned_buffer_size(self):
        in_data = b"This is a text that doesn't quite fit in a single 16-byte chunk."
        f = AsyncGridIn(self.db.fs, chunkSize=16)
        await f.write(in_data)
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        out_data = b""
        while 1:
            s = await g.read(13)
            if not s:
                break
            out_data += s

        self.assertEqual(in_data, out_data)

    async def test_readchunk(self):
        in_data = b"a" * 10
        f = AsyncGridIn(self.db.fs, chunkSize=3)
        await f.write(in_data)
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(3, len(await g.readchunk()))

        self.assertEqual(2, len(await g.read(2)))
        self.assertEqual(1, len(await g.readchunk()))

        self.assertEqual(3, len(await g.read(3)))

        self.assertEqual(1, len(await g.readchunk()))

        self.assertEqual(0, len(await g.readchunk()))

    async def test_write_unicode(self):
        f = AsyncGridIn(self.db.fs)
        with self.assertRaises(TypeError):
            await f.write("foo")

        f = AsyncGridIn(self.db.fs, encoding="utf-8")
        await f.write("foo")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual(b"foo", await g.read())

        f = AsyncGridIn(self.db.fs, encoding="iso-8859-1")
        await f.write("aé")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)
        self.assertEqual("aé".encode("iso-8859-1"), await g.read())

    async def test_set_after_close(self):
        f = AsyncGridIn(self.db.fs, _id="foo", bar="baz")

        self.assertEqual("foo", f._id)
        self.assertEqual("baz", f.bar)
        self.assertRaises(AttributeError, getattr, f, "baz")
        self.assertRaises(AttributeError, getattr, f, "uploadDate")

        self.assertRaises(AttributeError, setattr, f, "_id", 5)
        if _IS_SYNC:
            f.bar = "foo"
            f.baz = 5
        else:
            await f.set("bar", "foo")
            await f.set("baz", 5)

        self.assertEqual("foo", f._id)
        self.assertEqual("foo", f.bar)
        self.assertEqual(5, f.baz)
        self.assertRaises(AttributeError, getattr, f, "uploadDate")

        await f.close()

        self.assertEqual("foo", f._id)
        self.assertEqual("foo", f.bar)
        self.assertEqual(5, f.baz)
        self.assertTrue(f.uploadDate)

        self.assertRaises(AttributeError, setattr, f, "_id", 5)
        if _IS_SYNC:
            f.bar = "a"
            f.baz = "b"
        else:
            await f.set("bar", "a")
            await f.set("baz", "b")
        self.assertRaises(AttributeError, setattr, f, "upload_date", 5)

        g = AsyncGridOut(self.db.fs, f._id)
        if not _IS_SYNC:
            await g.open()
        self.assertEqual("a", g.bar)
        self.assertEqual("b", g.baz)
        # Versions 2.0.1 and older saved a _closed field for some reason.
        self.assertRaises(AttributeError, getattr, g, "_closed")

    async def test_context_manager(self):
        contents = b"Imagine this is some important data..."

        async with AsyncGridIn(self.db.fs, filename="important") as infile:
            await infile.write(contents)

        async with AsyncGridOut(self.db.fs, infile._id) as outfile:
            self.assertEqual(contents, await outfile.read())

    async def test_exception_file_non_existence(self):
        contents = b"Imagine this is some important data..."

        with self.assertRaises(ConnectionError):
            async with AsyncGridIn(self.db.fs, filename="important") as infile:
                await infile.write(contents)
                raise ConnectionError("Test exception")

        # Expectation: File chunks are written, entry in files doesn't appear.
        self.assertEqual(
            await self.db.fs.chunks.count_documents({"files_id": infile._id}), infile._chunk_number
        )

        self.assertIsNone(await self.db.fs.files.find_one({"_id": infile._id}))
        self.assertTrue(infile.closed)

    async def test_prechunked_string(self):
        async def write_me(s, chunk_size):
            buf = BytesIO(s)
            infile = AsyncGridIn(self.db.fs)
            while True:
                to_write = buf.read(chunk_size)
                if to_write == b"":
                    break
                await infile.write(to_write)
            await infile.close()
            buf.close()

            outfile = AsyncGridOut(self.db.fs, infile._id)
            data = await outfile.read()
            self.assertEqual(s, data)

        s = b"x" * DEFAULT_CHUNK_SIZE * 4
        # Test with default chunk size
        await write_me(s, DEFAULT_CHUNK_SIZE)
        # Multiple
        await write_me(s, DEFAULT_CHUNK_SIZE * 3)
        # Custom
        await write_me(s, 262300)

    async def test_grid_out_lazy_connect(self):
        fs = self.db.fs
        outfile = AsyncGridOut(fs, file_id=-1)
        with self.assertRaises(NoFile):
            await outfile.read()
        with self.assertRaises(NoFile):
            if not _IS_SYNC:
                await outfile.open()
            outfile.filename

        infile = AsyncGridIn(fs, filename=1)
        await infile.close()

        outfile = AsyncGridOut(fs, infile._id)
        await outfile.read()
        outfile.filename

        outfile = AsyncGridOut(fs, infile._id)
        await outfile.readchunk()

    async def test_grid_in_lazy_connect(self):
        client = self.simple_client("badhost", connect=False, serverSelectionTimeoutMS=10)
        fs = client.db.fs
        infile = AsyncGridIn(fs, file_id=-1, chunk_size=1)
        with self.assertRaises(ServerSelectionTimeoutError):
            await infile.write(b"data")
        with self.assertRaises(ServerSelectionTimeoutError):
            await infile.close()

    async def test_unacknowledged(self):
        # w=0 is prohibited.
        with self.assertRaises(ConfigurationError):
            AsyncGridIn((await self.async_rs_or_single_client(w=0)).pymongo_test.fs)

    async def test_survive_cursor_not_found(self):
        # By default the find command returns 101 documents in the first batch.
        # Use 102 batches to cause a single getMore.
        chunk_size = 1024
        data = b"d" * (102 * chunk_size)
        listener = OvertCommandListener()
        client = await self.async_rs_or_single_client(event_listeners=[listener])
        db = client.pymongo_test
        async with AsyncGridIn(db.fs, chunk_size=chunk_size) as infile:
            await infile.write(data)

        async with AsyncGridOut(db.fs, infile._id) as outfile:
            self.assertEqual(len(await outfile.readchunk()), chunk_size)

            # Kill the cursor to simulate the cursor timing out on the server
            # when an application spends a long time between two calls to
            # readchunk().
            assert await client.address is not None
            await client._close_cursor_now(
                outfile._chunk_iter._cursor.cursor_id,
                _CursorAddress(await client.address, db.fs.chunks.full_name),  # type: ignore[arg-type]
            )

            # Read the rest of the file without error.
            self.assertEqual(len(await outfile.read()), len(data) - chunk_size)

        # Paranoid, ensure that a getMore was actually sent.
        self.assertIn("getMore", listener.started_command_names())

    @async_client_context.require_sync
    async def test_zip(self):
        zf = BytesIO()
        z = zipfile.ZipFile(zf, "w")
        z.writestr("test.txt", b"hello world")
        z.close()
        zf.seek(0)

        f = AsyncGridIn(self.db.fs, filename="test.zip")
        await f.write(zf)
        await f.close()
        self.assertEqual(1, await self.db.fs.files.count_documents({}))
        self.assertEqual(1, await self.db.fs.chunks.count_documents({}))

        g = AsyncGridOut(self.db.fs, f._id)
        z = zipfile.ZipFile(g)
        self.assertSequenceEqual(z.namelist(), ["test.txt"])
        self.assertEqual(z.read("test.txt"), b"hello world")

    async def test_grid_out_unsupported_operations(self):
        f = AsyncGridIn(self.db.fs, chunkSize=3)
        await f.write(b"hello world")
        await f.close()

        g = AsyncGridOut(self.db.fs, f._id)

        self.assertRaises(io.UnsupportedOperation, g.writelines, [b"some", b"lines"])
        self.assertRaises(io.UnsupportedOperation, g.write, b"some text")
        self.assertRaises(io.UnsupportedOperation, g.fileno)
        self.assertRaises(io.UnsupportedOperation, g.truncate)

        self.assertFalse(g.writable())
        self.assertFalse(g.isatty())


if __name__ == "__main__":
    unittest.main()



# Copyright 2015-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for representing raw BSON documents.

Inserting and Retrieving RawBSONDocuments
=========================================

Example: Moving a document between different databases/collections

.. doctest::

  >>> import bson
  >>> from pymongo import MongoClient
  >>> from bson.raw_bson import RawBSONDocument
  >>> client = MongoClient(document_class=RawBSONDocument)
  >>> client.drop_database("db")
  >>> client.drop_database("replica_db")
  >>> db = client.db
  >>> result = db.test.insert_many(
  ...     [{"_id": 1, "a": 1}, {"_id": 2, "b": 1}, {"_id": 3, "c": 1}, {"_id": 4, "d": 1}]
  ... )
  >>> replica_db = client.replica_db
  >>> for doc in db.test.find():
  ...     print(f"raw document: {doc.raw}")
  ...     print(f"decoded document: {bson.decode(doc.raw)}")
  ...     result = replica_db.test.insert_one(doc)
  ...
  raw document: b'...'
  decoded document: {'_id': 1, 'a': 1}
  raw document: b'...'
  decoded document: {'_id': 2, 'b': 1}
  raw document: b'...'
  decoded document: {'_id': 3, 'c': 1}
  raw document: b'...'
  decoded document: {'_id': 4, 'd': 1}

For use cases like moving documents across different databases or writing binary
blobs to disk, using raw BSON documents provides better speed and avoids the
overhead of decoding or encoding BSON.
"""
from __future__ import annotations

from typing import Any, ItemsView, Iterator, Mapping, Optional

from bson import _get_object_size, _raw_to_dict
from bson.codec_options import _RAW_BSON_DOCUMENT_MARKER, CodecOptions
from bson.codec_options import DEFAULT_CODEC_OPTIONS as DEFAULT


def _inflate_bson(
    bson_bytes: bytes | memoryview,
    codec_options: CodecOptions[RawBSONDocument],
    raw_array: bool = False,
) -> dict[str, Any]:
    """Inflates the top level fields of a BSON document.

    :param bson_bytes: the BSON bytes that compose this document
    :param codec_options: An instance of
        :class:`~bson.codec_options.CodecOptions` whose ``document_class``
        must be :class:`RawBSONDocument`.
    """
    return _raw_to_dict(bson_bytes, 4, len(bson_bytes) - 1, codec_options, {}, raw_array=raw_array)


class RawBSONDocument(Mapping[str, Any]):
    """Representation for a MongoDB document that provides access to the raw
    BSON bytes that compose it.

    Only when a field is accessed or modified within the document does
    RawBSONDocument decode its bytes.
    """

    __slots__ = ("__raw", "__inflated_doc", "__codec_options")
    _type_marker = _RAW_BSON_DOCUMENT_MARKER
    __codec_options: CodecOptions[RawBSONDocument]

    def __init__(
        self,
        bson_bytes: bytes | memoryview,
        codec_options: Optional[CodecOptions[RawBSONDocument]] = None,
    ) -> None:
        """Create a new :class:`RawBSONDocument`

        :class:`RawBSONDocument` is a representation of a BSON document that
        provides access to the underlying raw BSON bytes. Only when a field is
        accessed or modified within the document does RawBSONDocument decode
        its bytes.

        :class:`RawBSONDocument` implements the ``Mapping`` abstract base
        class from the standard library so it can be used like a read-only
        ``dict``::

            >>> from bson import encode
            >>> raw_doc = RawBSONDocument(encode({'_id': 'my_doc'}))
            >>> raw_doc.raw
            b'...'
            >>> raw_doc['_id']
            'my_doc'

        :param bson_bytes: the BSON bytes that compose this document
        :param codec_options: An instance of
            :class:`~bson.codec_options.CodecOptions` whose ``document_class``
            must be :class:`RawBSONDocument`. The default is
            :attr:`DEFAULT_RAW_BSON_OPTIONS`.

        .. versionchanged:: 3.8
          :class:`RawBSONDocument` now validates that the ``bson_bytes``
          passed in represent a single bson document.

        .. versionchanged:: 3.5
          If a :class:`~bson.codec_options.CodecOptions` is passed in, its
          `document_class` must be :class:`RawBSONDocument`.
        """
        self.__raw = bson_bytes
        self.__inflated_doc: Optional[Mapping[str, Any]] = None
        # Can't default codec_options to DEFAULT_RAW_BSON_OPTIONS in signature,
        # it refers to this class RawBSONDocument.
        if codec_options is None:
            codec_options = DEFAULT_RAW_BSON_OPTIONS
        elif not issubclass(codec_options.document_class, RawBSONDocument):
            raise TypeError(
                "RawBSONDocument cannot use CodecOptions with document "
                f"class {codec_options.document_class}"
            )
        self.__codec_options = codec_options
        # Validate the bson object size.
        _get_object_size(bson_bytes, 0, len(bson_bytes))

    @property
    def raw(self) -> bytes | memoryview:
        """The raw BSON bytes composing this document."""
        return self.__raw

    def items(self) -> ItemsView[str, Any]:
        """Lazily decode and iterate elements in this document."""
        return self.__inflated.items()

    @property
    def __inflated(self) -> Mapping[str, Any]:
        if self.__inflated_doc is None:
            # We already validated the object's size when this document was
            # created, so no need to do that again.
            self.__inflated_doc = self._inflate_bson(self.__raw, self.__codec_options)
        return self.__inflated_doc

    @staticmethod
    def _inflate_bson(
        bson_bytes: bytes | memoryview, codec_options: CodecOptions[RawBSONDocument]
    ) -> Mapping[str, Any]:
        return _inflate_bson(bson_bytes, codec_options)

    def __getitem__(self, item: str) -> Any:
        return self.__inflated[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self.__inflated)

    def __len__(self) -> int:
        return len(self.__inflated)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RawBSONDocument):
            return self.__raw == other.raw
        return NotImplemented

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.raw!r}, codec_options={self.__codec_options!r})"


class _RawArrayBSONDocument(RawBSONDocument):
    """A RawBSONDocument that only expands sub-documents and arrays when accessed."""

    @staticmethod
    def _inflate_bson(
        bson_bytes: bytes | memoryview, codec_options: CodecOptions[RawBSONDocument]
    ) -> Mapping[str, Any]:
        return _inflate_bson(bson_bytes, codec_options, raw_array=True)


DEFAULT_RAW_BSON_OPTIONS: CodecOptions[RawBSONDocument] = DEFAULT.with_options(
    document_class=RawBSONDocument
)
_RAW_ARRAY_BSON_OPTIONS: CodecOptions[_RawArrayBSONDocument] = DEFAULT.with_options(
    document_class=_RawArrayBSONDocument
)
"""The default :class:`~bson.codec_options.CodecOptions` for
:class:`RawBSONDocument`.
"""



# /// script
# dependencies = [
#   "uvloop>=0.18"
# ]
# requires-python = ">=3.10"
# ///
from __future__ import annotations

import sys
from pathlib import Path

import uvloop

# Use pymongo from parent directory.
root = Path(__file__).parent.parent
sys.path.insert(0, str(root))

from test.asynchronous import async_simple_test_client  # noqa: E402


async def main():
    async with async_simple_test_client() as client:
        result = await client.admin.command("ping")
        assert result["ok"]


uvloop.run(main())



# Copyright 2023-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
from test import IntegrationTest, client_context, unittest
from unittest.mock import patch

from bson import json_util
from pymongo.errors import OperationFailure
from pymongo.logger import _DEFAULT_DOCUMENT_LENGTH

_IS_SYNC = True


# https://github.com/mongodb/specifications/tree/master/source/command-logging-and-monitoring/tests#prose-tests
class TestLogger(IntegrationTest):
    def test_default_truncation_limit(self):
        docs = [{"x": "y"} for _ in range(100)]
        db = self.db

        with patch.dict("os.environ"):
            os.environ.pop("MONGOB_LOG_MAX_DOCUMENT_LENGTH", None)
            with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                db.test.insert_many(docs)

                cmd_started_log = json_util.loads(cm.records[0].getMessage())
                self.assertEqual(len(cmd_started_log["command"]), _DEFAULT_DOCUMENT_LENGTH + 3)

                cmd_succeeded_log = json_util.loads(cm.records[1].getMessage())
                self.assertLessEqual(len(cmd_succeeded_log["reply"]), _DEFAULT_DOCUMENT_LENGTH + 3)

            with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                db.test.find({}).to_list()
                cmd_succeeded_log = json_util.loads(cm.records[1].getMessage())
                self.assertEqual(len(cmd_succeeded_log["reply"]), _DEFAULT_DOCUMENT_LENGTH + 3)

    def test_configured_truncation_limit(self):
        cmd = {"hello": True}
        db = self.db
        with patch.dict("os.environ", {"MONGOB_LOG_MAX_DOCUMENT_LENGTH": "5"}):
            with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                db.command(cmd)

                cmd_started_log = json_util.loads(cm.records[0].getMessage())
                self.assertEqual(len(cmd_started_log["command"]), 5 + 3)

                cmd_succeeded_log = json_util.loads(cm.records[1].getMessage())
                self.assertLessEqual(len(cmd_succeeded_log["reply"]), 5 + 3)
                with self.assertRaises(OperationFailure):
                    db.command({"notARealCommand": True})
                cmd_failed_log = json_util.loads(cm.records[-1].getMessage())
                self.assertEqual(len(cmd_failed_log["failure"]), 5 + 3)

    def test_truncation_multi_byte_codepoints(self):
        document_lengths = ["20000", "20001", "20002"]
        multi_byte_char_str_len = 50_000
        str_to_repeat = "界"

        multi_byte_char_str = ""
        for i in range(multi_byte_char_str_len):
            multi_byte_char_str += str_to_repeat

        for length in document_lengths:
            with patch.dict("os.environ", {"MONGOB_LOG_MAX_DOCUMENT_LENGTH": length}):
                with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                    self.db.test.insert_one({"x": multi_byte_char_str})
                    cmd_started_log = json_util.loads(cm.records[0].getMessage())["command"]

                    cmd_started_log = cmd_started_log[:-3]
                    last_3_bytes = cmd_started_log.encode()[-3:].decode()

                    self.assertEqual(last_3_bytes, str_to_repeat)

    def test_logging_without_listeners(self):
        c = self.single_client()
        self.assertEqual(len(c._event_listeners.event_listeners()), 0)
        with self.assertLogs("pymongo.connection", level="DEBUG") as cm:
            c.db.test.insert_one({"x": "1"})
            self.assertGreater(len(cm.records), 0)
        with self.assertLogs("pymongo.command", level="DEBUG") as cm:
            c.db.test.insert_one({"x": "1"})
            self.assertGreater(len(cm.records), 0)
        with self.assertLogs("pymongo.serverSelection", level="DEBUG") as cm:
            c.db.test.insert_one({"x": "1"})
            self.assertGreater(len(cm.records), 0)

    @client_context.require_failCommand_fail_point
    def test_logging_retry_read_attempts(self):
        self.db.test.insert_one({"x": "1"})

        with self.fail_point(
            {
                "mode": {"times": 1},
                "data": {
                    "failCommands": ["find"],
                    "errorCode": 10107,
                    "errorLabels": ["RetryableWriteError"],
                },
            }
        ):
            with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                self.db.test.find_one({"x": "1"})

        retry_messages = [
            r.getMessage() for r in cm.records if "Retrying read attempt" in r.getMessage()
        ]
        self.assertEqual(len(retry_messages), 1)

    @client_context.require_failCommand_fail_point
    @client_context.require_retryable_writes
    def test_logging_retry_write_attempts(self):
        with self.fail_point(
            {
                "mode": {"times": 1},
                "data": {
                    "errorCode": 10107,
                    "errorLabels": ["RetryableWriteError"],
                    "failCommands": ["insert"],
                },
            }
        ):
            with self.assertLogs("pymongo.command", level="DEBUG") as cm:
                self.db.test.insert_one({"x": "1"})

        retry_messages = [
            r.getMessage() for r in cm.records if "Retrying write attempt" in r.getMessage()
        ]
        self.assertEqual(len(retry_messages), 1)


if __name__ == "__main__":
    unittest.main()
